@@ -28,7 +28,7 @@ import numpy as np
 
 # Local imports...
 from .codestream import Codestream
-from . import core, version
+from . import core, version, get_option
 from .jp2box import (Jp2kBox, JPEG2000SignatureBox, FileTypeBox,
                      JP2HeaderBox, ColourSpecificationBox,
                      ContiguousCodestreamBox, ImageHeaderBox)
@@ -51,6 +51,8 @@ class Jp2k(Jp2kBox):
 
     Properties
     ----------
+    codestream : glymur.codestream.Codestream
+        JP2 or J2K codestream object.
     ignore_pclr_cmap_cdef : bool
         Whether or not to ignore the pclr, cmap, or cdef boxes during any
         color transformation, defaults to False.
@@ -59,8 +61,6 @@ class Jp2k(Jp2kBox):
     verbose : bool
         Whether or not to print informational messages produced by the
         OpenJPEG library, defaults to false.
-    codestream : glymur.codestream.Codestream
-        JP2 or J2K codestream object.
 
     Examples
     --------
@@ -73,11 +73,25 @@ class Jp2k(Jp2kBox):
     >>> image.shape
     (1456, 2592, 3)
 
-    Read a lower resolution thumbnail.
+    Read a lower resolution thumbnail
 
     >>> thumbnail = jp2[::2, ::2]
     >>> thumbnail.shape
     (728, 1296, 3)
+
+    Make use of OpenJPEG's thread support
+    >>> import glymur
+    >>> import time
+    >>> if glymur.version.openjpeg_version >= '2.2.0':
+    ...     jp2file = glymur.data.nemo()
+    ...     jp2 = glymur.Jp2k(jp2file)
+    ...     t0 = time.time(); data = jp2[:]; t1 = time.time()
+    ...     t1 - t0 #doctest: +SKIP
+    0.9024193286895752
+    ...     glymur.set_options('lib.num_threads', 4)
+    ...     t0 = time.time(); data = jp2[:]; t1 = time.time()
+    ...     t1 - t0 #doctest: +SKIP
+    0.4060473537445068
     """
 
     def __init__(self, filename, data=None, shape=None, **kwargs):
@@ -140,7 +154,7 @@ class Jp2k(Jp2kBox):
         """
         Jp2kBox.__init__(self)
 
-        # In case of pathlib.Paths... 
+        # In case of pathlib.Paths...
         self.filename = str(filename)
 
         self.box = []
@@ -392,9 +406,22 @@ class Jp2k(Jp2kBox):
                    "options.")
             raise IOError(msg)
 
-        if cratios is not None and psnr is not None:
-            msg = "Cannot specify cratios and psnr options together."
-            raise IOError(msg)
+        if psnr is not None:
+            if cratios is not None:
+                msg = "Cannot specify cratios and psnr options together."
+                raise IOError(msg)
+
+            if 0 in psnr and psnr[-1] != 0:
+                msg = ("If a zero value is supplied in the PSNR keyword "
+                       "argument, it must be in the final position.")
+                raise IOError(msg)
+
+            if (((0 in psnr and np.any(np.diff(psnr[:-1]) < 0)) or
+                 (0 not in psnr and np.any(np.diff(psnr) < 0)))):
+                msg = ("PSNR values must be increasing, with one exception - "
+                       "zero may be in the final position to indicate a "
+                       "lossless layer.")
+                raise IOError(msg)
 
         if version.openjpeg_version_tuple[0] == 1:
             cparams = opj.set_default_encoder_parameters()
@@ -411,22 +438,15 @@ class Jp2k(Jp2kBox):
         else:
             cparams.codec_fmt = opj2.CODEC_J2K
 
-        # Set defaults to lossless to begin.
-        cparams.tcp_rates[0] = 0
-        cparams.tcp_numlayers = 1
-        cparams.cp_disto_alloc = 1
-
         cparams.irreversible = 1 if irreversible else 0
 
         if cinema2k is not None:
             self._cparams = cparams
             self._set_cinema_params('cinema2k', cinema2k)
-            return
 
         if cinema4k is not None:
             self._cparams = cparams
             self._set_cinema_params('cinema4k', cinema4k)
-            return
 
         if cbsize is not None:
             cparams.cblockw_init = cbsize[1]
@@ -490,6 +510,12 @@ class Jp2k(Jp2kBox):
                 raise IOError(msg)
             cparams.tcp_mct = 1 if mct else 0
 
+        # Set defaults to lossless to begin.
+        if cparams.tcp_numlayers == 0:
+            cparams.tcp_rates[0] = 0
+            cparams.tcp_numlayers += 1
+            cparams.cp_disto_alloc = 1
+
         self._validate_compression_params(img_array, cparams, colorspace)
 
         self._cparams = cparams
@@ -1253,7 +1279,6 @@ class Jp2k(Jp2kBox):
         RuntimeError
             If the image has differing subsample factors.
         """
-        self.layer = layer
         self._subsampling_sanity_check()
         self._populate_dparams(rlevel, tile=tile, area=area)
         image = self._read_openjp2_common()
@@ -1285,6 +1310,9 @@ class Jp2k(Jp2kBox):
                 opj2.set_info_handler(codec, None)
 
             opj2.setup_decoder(codec, self._dparams)
+            if version.openjpeg_version >= '2.2.0':
+                opj2.codec_set_threads(codec, get_option('lib.num_threads'))
+
             raw_image = opj2.read_header(stream, codec)
             stack.callback(opj2.image_destroy, raw_image)
 
@@ -1500,9 +1528,9 @@ class Jp2k(Jp2kBox):
             raise IOError(msg)
 
         if component.sgnd:
-            dtype = np.int8 if component.prec <=8 else np.int16
+            dtype = np.int8 if component.prec <= 8 else np.int16
         else:
-            dtype = np.uint8 if component.prec <=8 else np.uint16
+            dtype = np.uint8 if component.prec <= 8 else np.uint16
 
         return dtype
 
@@ -1888,6 +1916,7 @@ def _default_warning_handler(library_msg, _):
     msg = "OpenJPEG library warning:  {0}".format(library_msg)
     warnings.warn(msg, UserWarning)
 
+
 _ERROR_CALLBACK = _CMPFUNC(_default_error_handler)
 _INFO_CALLBACK = _CMPFUNC(_default_info_handler)
 _WARNING_CALLBACK = _CMPFUNC(_default_warning_handler)
